from enum import Enum
from typing import Dict, List, Optional

from pydantic import BaseModel, validator


class DbtResourceType(str, Enum):
    model = 'model'
    test = 'test'
    snapshot = 'snapshot'
    seed = 'seed'

    # dbt resource types we are not interested in, but we still need them in order for
    # dbt-airflow to be able to parse the manifest file
    operation = 'operation'


class NodeDeps(BaseModel):
    nodes: List[str]

    def __getitem__(self, item):
        return getattr(self, item)

    @validator('nodes')
    def filter(cls, val):
        """
        Filters out dbt tests since these will be constructed by subsequent steps
        """
        return [
            n for n in val
            if n.split('.')[0] in [
                DbtResourceType.model, DbtResourceType.seed, DbtResourceType.snapshot
            ]
        ]


class Node(BaseModel):
    resource_type: DbtResourceType
    depends_on: Optional[NodeDeps]
    fqn: Optional[List[str]]
    package_name: str
    name: str
    task_group: Optional[str] = None

    @validator('task_group', always=True)
    def create_task_group(cls, v, values) -> str:
        tg_idx = -2
        if values['resource_type'] == DbtResourceType.snapshot:
            tg_idx = -3

        if len(values['fqn']) >= abs(tg_idx):
            return values['fqn'][tg_idx]
        return values['fqn'][0]


class Manifest(BaseModel):
    nodes: Dict[str, Node]

    @validator('nodes')
    def filter(cls, val):
        return {
            k: v for k, v in val.items()
            if v.resource_type.value in ('model', 'seed', 'snapshot', 'test')
        }

    def get_statistics(self):
        """
        Returns a dictionary containing some statistics of the input manifest data.
        """
        node_types = [n.resource_type for n in self.nodes.values()]
        return {
            'models': node_types.count(DbtResourceType.model),
            'tests': node_types.count(DbtResourceType.test),
            'snapshots': node_types.count(DbtResourceType.snapshot),
            'seeds': node_types.count(DbtResourceType.seed),
        }



class DbtAirflowException(Exception):
    pass


class ManifestNotFound(DbtAirflowException):
    """Raised when test_manifest.json file is not found under the expected path"""
    pass


class ManifestDataNotFound(DbtAirflowException):
    """Raised when a ManifestProcessor instance has empty data"""
    pass


class DbtCommandNotSupported(DbtAirflowException):
    """Raised when the given dbt command is not supported by dbt-airflow"""
    pass


class InvalidDbtCommand(DbtAirflowException):
    """Raised when the given task has an unexpected dbt command"""
    pass


class DuplicateTaskName(DbtAirflowException):
    """Raised when a duplicate task is identified within a TaskList"""
    pass


class OperatorClassNotSupported(DbtAirflowException):
    """Raised when an unsupported operator_class is provided in DbtTaskGroup"""
    pass



from pathlib import Path
from typing import Any, Optional

from airflow.operators.bash import BashOperator
from airflow.utils.decorators import apply_defaults


class DbtBashOperator(BashOperator):
    """
    This child class inherits from BashOperator and enriches its functionality such that various
    dbt commands can be executed as Airflow Operators. Given that dbt is a command-line tool, it
    makes sense to run these commands from bash (and thus BashOperator is used in the context of
    Airflow).
    """
    @apply_defaults
    def __init__(
        self,
        dbt_target_profile: str,
        dbt_profile_path: Path,
        dbt_project_path: Path,
        resource_name: str,
        dbt_command: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.dbt_target_profile = dbt_target_profile
        self.dbt_profile_path = dbt_profile_path
        self.dbt_project_path = dbt_project_path
        self.resource_name = resource_name
        self.dbt_command = dbt_command

        super().__init__(bash_command=self._get_bash_command(), **kwargs)

    def _get_bash_command(self) -> str:
        """
        Returns the bash command to be executed from BashOperator parent class on Airflow.
        We specify the `--no-write-json` flag in order to avoid overwriting the `manifest.json`
        file that sits at the heart of this implementation. Therefore, the creation and versioning
        of that file remains responsibility of the user running the library.
        """
        return f'cd {self.dbt_project_path.as_posix()} && ' \
               f'dbt --no-write-json ' \
               f'{self.dbt_command} ' \
               f'--profiles-dir {self.dbt_profile_path.as_posix()} ' \
               f'--target {self.dbt_target_profile} ' \
               f'--select {self.resource_name}'


class DbtRunBashOperator(DbtBashOperator):
    """
    Class for an Airflow Operator that executes a dbt `run` operation.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dbt_command='run', **kwargs)


class DbtSeedBashOperator(DbtBashOperator):
    """
    Class for an Airflow Operator that executes a dbt `seed` operation.
    """
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dbt_command='seed', **kwargs)


class DbtTestBashOperator(DbtBashOperator):
    """
    Class for an Airflow Operator that executes a dbt `test` operation.
    """
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dbt_command='test', **kwargs)


class DbtSnapshotBashOperator(DbtBashOperator):
    """
    Class for an Airflow Operator that executes a dbt `snapshot` operation.
    """
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dbt_command='snapshot', **kwargs)



from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Optional, Set

from airflow.models.baseoperator import BaseOperator

from dbt_airflow.parser.dbt import DbtResourceType, Node
from dbt_airflow.operators.bash import (
    DbtRunBashOperator,
    DbtTestBashOperator,
    DbtSeedBashOperator,
    DbtSnapshotBashOperator,
)
from dbt_airflow.operators.kubernetes import (
    DbtRunKubernetesPodOperator,
    DbtTestKubernetesPodOperator,
    DbtSeedKubernetesPodOperator,
    DbtSnapshotKubernetesPodOperator,
)


@dataclass
class AirflowTask:
    task_id: str

    def __eq__(self, other):
        return self.task_id == other.task_id


@dataclass(eq=False)
class ExtraTask(AirflowTask):
    """
    This is a dataclass representing an extra task that is supposed to be inserted in a specific
    place within the populated dbt project, based on the task names specified in
    `upstream_task_ids` and/or `downstream_task_ids`.
    """
    operator: BaseOperator
    operator_args: Dict[Any, Any] = field(default_factory=dict)
    upstream_task_ids: Set[str] = field(default_factory=set)
    downstream_task_ids: Set[str] = field(default_factory=set)
    task_group: Optional[str] = None


@dataclass(eq=False)
class DbtAirflowTask(AirflowTask):
    dbt_operator: Callable = field(init=False)
    manifest_node_name: str  # This is the node name as found on manifest.json file
    model_name: str = field(init=False)
    package_name: str
    resource_type: DbtResourceType
    task_group: Optional[str]
    upstream_task_ids: Set[str]
    operator_class: str

    def __post_init__(self):
        self.dbt_operator = self.get_dbt_operator()
        self.model_name = self.get_model_name()

    @classmethod
    def from_manifest_node(cls, manifest_node_name: str, node: Node, operator_class: str):
        """
        Creates an instance of this dataclass from a Manifest Node
        """
        return DbtAirflowTask(
            task_id=manifest_node_name,
            manifest_node_name=manifest_node_name,
            resource_type=node.resource_type,
            upstream_task_ids=set(node.depends_on.nodes),
            task_group=node.task_group,
            package_name=node.package_name,
            operator_class=operator_class,
        )

    @classmethod
    def test_task_from_manifest_node(cls, parent_manifest_node_name: str, parent_node: Node, operator_class: str):
        """
        Creates an instance of this dataclass that corresponds to a dbt task, from the input
        parent node.
        """
        model_name = parent_manifest_node_name.split('.')[-1]
        test_node_name = f'test.{parent_node.package_name}.{model_name}'
        return DbtAirflowTask(
            task_id=test_node_name,
            manifest_node_name='',
            resource_type=DbtResourceType.test,
            upstream_task_ids={parent_manifest_node_name},
            task_group=parent_node.task_group,
            package_name=parent_node.package_name,
            operator_class=operator_class,
        )

    def get_dbt_operator(self) -> Callable:
        """
        Returns a callable that corresponds to the dbt Airflow Operator based on the instance's
        resource type.
        """
        if self.operator_class == 'BashOperator':
            if self.resource_type == DbtResourceType.model:
                return DbtRunBashOperator
            if self.resource_type == DbtResourceType.test:
                return DbtTestBashOperator
            if self.resource_type == DbtResourceType.seed:
                return DbtSeedBashOperator
            return DbtSnapshotBashOperator

        if self.operator_class == 'KubernetesPodOperator':
            if self.resource_type == DbtResourceType.model:
                return DbtRunKubernetesPodOperator
            if self.resource_type == DbtResourceType.test:
                return DbtTestKubernetesPodOperator
            if self.resource_type == DbtResourceType.seed:
                return DbtSeedKubernetesPodOperator
            return DbtSnapshotKubernetesPodOperator

    def get_model_name(self) -> str:
        """
        Extracts the model name from the manifest node name (in the form
        `<resource-type>.<package-name>.<model-name>`) of the dbt resource.
        The test tasks get created by dbt-airflow and they all have an empty string as
        `manifest_node_name`. Therefore, the model name is extracted by the upstream dependencies
        of the test task, that always contains a single model/seed/snapshot task.
        """
        if self.resource_type != DbtResourceType.test:
            return self.manifest_node_name.split('.')[-1]
        return list(self.upstream_task_ids)[0].split('.')[-1]



from pathlib import Path
from typing import Any, Dict, List, Optional

from airflow.utils.task_group import TaskGroup

from dbt_airflow.core.task import ExtraTask, DbtAirflowTask
from dbt_airflow.core.task_builder import DbtAirflowTaskBuilder
from dbt_airflow.exceptions import OperatorClassNotSupported


class DbtTaskGroup(TaskGroup):

    SUPPORTED_OPERATORS = ['BashOperator', 'KubernetesPodOperator']

    def __init__(
        self,
        dbt_manifest_path: Path,
        dbt_target: str,
        dbt_profile_path: Path,
        dbt_project_path: Path,
        operator_class: Optional[str] = 'BashOperator',
        operator_kwargs: Optional[Dict[Any, Any]] = None,
        create_sub_task_groups: Optional[bool] = True,
        extra_tasks: Optional[List[ExtraTask]] = None,
        *args,
        **kwargs
    ):
        self.dbt_manifest_path = dbt_manifest_path
        self.dbt_target = dbt_target
        self.dbt_profile_path = dbt_profile_path
        self.dbt_project_path = dbt_project_path
        self.operator_class = operator_class
        self.create_sub_task_groups = create_sub_task_groups
        self.extra_tasks = extra_tasks

        if operator_kwargs:
            self.operator_kwargs = operator_kwargs
        else:
            self.operator_kwargs = {}

        self._input_validation()

        super().__init__(*args, **kwargs)

        with self as _self:
            task_builder = DbtAirflowTaskBuilder(
                manifest_path=dbt_manifest_path.as_posix(),
                extra_tasks=extra_tasks,
                operator_class=operator_class,
            )
            self.tasks = task_builder.build_tasks()
            self.nested_task_groups = self._build_nested_task_groups()
            self._build_airflow_tasks()

    def _build_nested_task_groups(self) -> Dict[str, TaskGroup]:
        """
        """
        if not self.create_sub_task_groups:
            return {}

        task_groups = {}
        for task in self.tasks:
            if task.task_group and task.task_group not in task_groups:
                task_groups[task.task_group] = TaskGroup(task.task_group)
        return task_groups

    def _build_airflow_tasks(self):
        airflow_tasks = {
            task.task_id: task.dbt_operator(
                task_id=task.task_id,
                dbt_target_profile=self.dbt_target,
                dbt_profile_path=self.dbt_profile_path,
                dbt_project_path=self.dbt_project_path,
                resource_name=task.model_name,
                task_group=self.nested_task_groups.get(task.task_group),
                **self.operator_kwargs,
            )
            for task in self.tasks if isinstance(task, DbtAirflowTask)
        }
        airflow_tasks.update({
            task.task_id: task.operator(task_id=task.task_id, **task.operator_args)
            for task in self.tasks if isinstance(task, ExtraTask)
        })

        # Handle dependencies
        for task in self.tasks:
            for upstream_task in task.upstream_task_ids:
                airflow_tasks[task.task_id] << airflow_tasks[upstream_task]

    def _input_validation(self):
        """
        Validates user input
        """
        if self.operator_class not in self.SUPPORTED_OPERATORS:
            raise OperatorClassNotSupported(
                f'{self.operator_class} is not supported. '
                f'Please choose one of {self.SUPPORTED_OPERATORS}'
            )



import json
import logging
import os
from typing import List, Optional

from ..exceptions import ManifestNotFound, ManifestDataNotFound
from dbt_airflow.core.task import DbtAirflowTask, ExtraTask
from dbt_airflow.core.task_list import TaskList
from dbt_airflow.parser.dbt import DbtResourceType, Manifest, Node


class DbtAirflowTaskBuilder:
    """
    This class implements functionality that is used to populate the dbt project as an Airflow DAG,
    based on the user input. It also handles dbt model and test dependencies as well as any extra
    tasks specified by user.
    """

    def __init__(
        self,
        manifest_path: str,
        operator_class: str,
        extra_tasks: Optional[List[ExtraTask]] = None,
    ) -> None:
        self.manifest_path = os.path.abspath(manifest_path)
        self.manifest = self._load_manifest()
        self.nodes_with_tests = set()
        self.task_list = TaskList()

        if extra_tasks is None:
            self.extra_tasks = []
        else:
            self.extra_tasks = extra_tasks

        self.operator_class = operator_class

    def _add_extra_tasks(self) -> None:
        """
        Adds extra tasks specified in DbtTaskGroup
        """
        if not self.extra_tasks:
            logging.info('No extra tasks were provided. Skipping..')
            return

        for task in self.extra_tasks:
            self.task_list.append(task)

            # Iterate through extra task downstream dependencies
            for downstream_task_id in task.downstream_task_ids:
                downstream_task = self.task_list.find_task_by_id(downstream_task_id)

                # We want to check if the downstream task of extra task, is defined as
                # upstream dependency
                for upstream_task_id in task.upstream_task_ids:
                    upstream_task = self.task_list.find_task_by_id(upstream_task_id)
                    if upstream_task.task_id in downstream_task.upstream_task_ids:
                        downstream_task.upstream_task_ids.remove(upstream_task_id)

                #
                downstream_task.upstream_task_ids.add(task.task_id)

    def _create_task(self, manifest_node_name: str, node: Node) -> None:
        """
        Creates a task from the input manifest node. If the node also has test, an additional
        test task also gets created.
        """
        self.task_list.append(
            DbtAirflowTask.from_manifest_node(
                manifest_node_name,
                node,
                self.operator_class,
            )
        )
        if manifest_node_name in self.nodes_with_tests:
            self.task_list.append(
                DbtAirflowTask.test_task_from_manifest_node(
                    manifest_node_name,
                    node,
                    self.operator_class,
                )
            )

    def _create_tasks_with_tests(self) -> None:
        """
        For every dbt model, seed and snapshot we call `_create_task()` method that is responsible
        for creating a task per resource as well as an additional test task in case at least
        one test was defined for that particular resource.
        """
        logging.info('Creating tasks from manifest.json file')
        for manifest_node_name, node in self.manifest.nodes.items():
            if node.resource_type in [
                DbtResourceType.model, DbtResourceType.seed, DbtResourceType.snapshot
            ]:
                self._create_task(manifest_node_name, node)

    def _get_nodes_with_tests(self) -> None:
        """
        Creates a set consisting of models, seeds and snapshots for which at least one test is
        found within the manifest file.
        """
        for node_name, node in self.manifest.nodes.items():
            if node.resource_type == DbtResourceType.test:
                self.nodes_with_tests.update(node.depends_on.nodes)
        logging.info(
            f'Found {len(self.nodes_with_tests)} models, snapshots or seeds with dbt tests.'
        )

    def _load_manifest(self) -> Manifest:
        """
        Loads the test_manifest.json file created by dbt
        """
        logging.info(f'Loading {self.manifest_path} file')
        if not os.path.isfile(self.manifest_path):
            raise ManifestNotFound(
                f'Manifest file was not found in {self.manifest_path}. '
                f'Make sure that you are running dbt-airflow from your dbt project directory '
                f'and that the project is compiled (hint: `dbt compile`).'
            )

        with open(self.manifest_path, 'r') as f:
            data = json.load(f)
            if not data:
                raise ManifestDataNotFound('No data was found.')

        manifest = Manifest(**data)
        logging.info(f'{self.manifest_path} file was loaded successfully.')
        logging.info(f'Found {manifest.get_statistics()}')

        return manifest

    def _update_dependencies(self) -> None:
        """
        Goes through every test task in the existing `self.task_list` and re-arranges the
        dependencies of the nodes such that:

        For every task that references test's parent task its upstream tasks, it will remove the
        dependency to test's parent task and add a new dependency that references the test task.

        This is required given that test tasks are not part of dbt manifest file. `dbt-airflow`
        creates a test task for every model, seed and snapshot resource that have at least one
        test defined. Therefore, we then need to fix the dependencies between all these tasks
        such that whenever a model/snapshot/seed task that also has at least one test defined
        and is referenced in upstream dependencies of other tasks, it is replaced by the
        corresponding test task.
        """
        for task in self.task_list:
            if task.resource_type == DbtResourceType.test:
                parent_task = self.task_list.find_task_by_id(list(task.upstream_task_ids)[0])
                for task_other in self.task_list:
                    if task_other.resource_type != DbtResourceType.test \
                            and parent_task.task_id in task_other.upstream_task_ids:
                        task_other.upstream_task_ids.remove(parent_task.task_id)
                        task_other.upstream_task_ids.add(task.task_id)

    def build_tasks(self) -> TaskList:
        """
        Algorithm steps:
            1. Firstly, we extract the model, seed and snapshot nodes that have tests
            2. We then start building the tasks for every model, seed and snapshot node
                 i. Create a task for the mode, seed or snapshot
                ii. If the node has a test, create a test task as well
            3. Now that we have the full set of tasks, we need to go back and re-arrange
                the dependencies since we have created test tasks for the nodes.
                We couldn't have done this step in step 2, since if we were about to fixing
                the upstream dependencies of other nodes referencing the model/seed/snapshot task
                on the fly, we would've missed some given that at every iteration of creating
                tasks we don't have the full list of tasks. Therefore, this needs to be done
                at the very end
        """
        logging.info('Building tasks from manifest data')

        # Step 1: Extract all models, seeds and snapshots that have at least one test
        self._get_nodes_with_tests()

        # Step 2: Create tasks with tests
        self._create_tasks_with_tests()

        # Step 3: Update dependencies between models and tests created
        self._update_dependencies()

        # Step 4: Add extra tasks
        self._add_extra_tasks()

        logging.info(f'Created a TaskList of: {self.task_list.get_statistics()}')
        logging.warning(
            'The number of tests created could be less of the original number of tests reported '
            'in manifest file. This is due to the fact that the resulting Airflow DAG will only '
            'create a single test task, to execute all the tests for a particular model, '
            'snapshot or seed.'
        )

        return self.task_list



import logging
from pathlib import Path
from typing import Any, Optional

from airflow.utils.decorators import apply_defaults

try:
    # apache-airflow-providers-cncf-kubernetes >= 7.4.0
    from airflow.providers.cncf.kubernetes.backcompat.backwards_compat_converters import (
        convert_env_vars,
    )
    from airflow.providers.cncf.kubernetes.operators.pod import KubernetesPodOperator
except ImportError:
    try:
        # apache-airflow-providers-cncf-kubernetes < 7.4.0
        from airflow.providers.cncf.kubernetes.operators.kubernetes_pod import (
            KubernetesPodOperator,
        )
    except ImportError as error:
        logging.exception(error)
        raise ImportError(
            'Could not import KubernetesPodOperator. Make sure you have '
            'apache-airflow-providers-cncf-kubernetes installed.'
        )


class DbtKubernetesPodOperator(KubernetesPodOperator):

    @apply_defaults
    def __init__(
        self,
        name: str,
        namespace: str,
        image: str,
        dbt_target_profile: str,
        dbt_profile_path: Path,
        dbt_project_path: Path,
        resource_name: str,
        dbt_command: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.name = name
        self.namespace = namespace
        self.image = image
        self.dbt_target_profile = dbt_target_profile
        self.dbt_profile_path = dbt_profile_path
        self.dbt_project_path = dbt_project_path
        self.resource_name = resource_name
        self.dbt_command = dbt_command

        super().__init__(
            name=self.name,
            namespace=self.namespace,
            image=self.image,
            cmds=self._get_cmds(),
            **kwargs,
        )

    def _get_cmds(self):
        return [
            'dbt', self.dbt_command,
            '--select', self.resource_name,
            '--project-dir', self.dbt_project_path.as_posix(),
            '--profiles-dir', self.dbt_profile_path.as_posix(),
            '--target', self.dbt_target_profile
        ]


class DbtRunKubernetesPodOperator(DbtKubernetesPodOperator):

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dbt_command='run', **kwargs)


class DbtTestKubernetesPodOperator(DbtKubernetesPodOperator):

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dbt_command='test', **kwargs)


class DbtSeedKubernetesPodOperator(DbtKubernetesPodOperator):

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dbt_command='seed', **kwargs)


class DbtSnapshotKubernetesPodOperator(DbtKubernetesPodOperator):

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dbt_command='snapshot', **kwargs)



from typing import Dict

from dbt_airflow.core.task import AirflowTask, DbtAirflowTask, ExtraTask
from dbt_airflow.parser.dbt import DbtResourceType


class TaskList(list):
    """
    A collection of AirflowTasks
    """

    def append(self, item) -> None:
        """
        Appends a new task into the `TaskList` assuming that every element is of type `AirflowTask`
        and no other task with the same `item.task_id` already exists in the TaskList.
        """
        if not isinstance(item, AirflowTask):
            raise ValueError(f'Element of type {type(item)} cannot be added in TaskList.')

        if item.task_id in [t.task_id for t in self]:
            raise ValueError(f'A task with id {item.task_id} already exists')

        super(TaskList, self).append(item)

    def find_task_by_id(self, task_id: str) -> AirflowTask:
        """
        Returns the task with the matching `task_id`. If no task is found a ValueError is raised
        """
        for task in self:
            if task.task_id == task_id:
                return task

        raise ValueError(f'Task with id `{task_id}` was not found.')

    def get_statistics(self) -> Dict[str, int]:
        """
        Returns counts per node type in the resulting TaskList
        """
        resource_types = [task.resource_type for task in self if isinstance(task, DbtAirflowTask)]
        return {
            'models': resource_types.count(DbtResourceType.model),
            'tests': resource_types.count(DbtResourceType.test),
            'snapshots': resource_types.count(DbtResourceType.snapshot),
            'seeds': resource_types.count(DbtResourceType.seed),
            'extra_tasks': sum(isinstance(task, ExtraTask) for task in self)
        }
